/*
 Adjustable 2A PSU
 Version: 1.0
 Author: Alex from insideGadgets (http://www.insidegadgets.com)
 Created: 16/10/2017
 Last Modified: 24/02/2018
 
 avrdude -c usbtiny -p atmega328p -U lfuse:w:0xff:m -U hfuse:w:0xd9:m -U efuse:w:0xff:m
 
  */

// ATmega328 QFP Pin Map
//
// 							PB7 XTAL					PD7 SW Enable (not used)
// PC6 Reset				PB6 XTAL			 		PD6 LED
// PC5 DPOT SDI SDO		PB5 SCK / DAC SDA		PD5 0 Ohm Enable
// PC4 SSD1306 CLK		PB4 MISO	/ DAC SCL	PD4 4.7 Ohm Enable
// PC3 SSD1306 MOSI		PB3 MOSI					PD3 10 Ohm Enable
// PC2 SSD1306 RST		PB2 SW Select			PD2 20 Ohm Enable
// PC1 SSD1306 DC			PB1 Encoder R			PD1 DPOT CS
// PC0 SSD1306 CS			PB0 Encoder L			PD0 DPOT SCK

#define F_CPU 12000000 // 12 MHz

#include <avr/io.h>
#include <avr/wdt.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <util/delay.h>
#include <avr/sleep.h>
#include <string.h>
#include "ssd1306.c"
#include "setup.c" // See defines, variables, constants, functions here

int main(void) {
	setup();
	
	while(1) {
		update_voltage_dpot_mosfet();
		update_current_limit();
		_delay_ms(10); // Wait a bit to adjust
		
		// 10 samples of voltage and current to average
		uint16_t adcVoltageAverage = 0;
		uint16_t adcCurrentAverage = 0;
		for (uint8_t x = 0; x < 10; x++) {
			adcVoltageAverage += adc_read(7);
			_delay_ms(1);
			adcCurrentAverage += adc_read(6);
			_delay_ms(1);
		}
		adcVoltageAverage = adcVoltageAverage / 10;
		adcCurrentAverage = adcCurrentAverage / 10;
		uint16_t adcVoltage = (int) (double) adcVoltageAverage * (double) 0.00107421875 * (double) 14.7 * (double) 100;
		uint16_t adcCurrent = (int) (double) adcCurrentAverage * (double) 0.00107421875 * (double) 2 * (double) 1000;
		
		// For display
		voltageNumber = adcVoltage;
		currentNumber = adcCurrent;
		
		
		// Calibration
		if (inCalibration == true && waitForLowVoltage == false) {
			if (adcVoltage > adcVoltageOld) { // If voltage is higher than old voltage, add to array
				adcVoltageOld = adcVoltage;
				
				PORTD |= (1<<LED_ENABLE);
				
				if (calibrationCounter < 400) {
					calibrationValues[calibrationCounter][0] = dpotValue;
					calibrationValues[calibrationCounter][1] = mosfetEnabled;
					calibrationCounter++;
				}
				else {
					inCalibration = false;
					calibrationLimit = calibrationCounter;
					calibrationCounter = 0;
					calibrated = true;
				}
			}
			else {
				PORTD &= ~(1<<LED_ENABLE);
			}
			
			// Increment dpot / mosfets if voltage doesn't drop by more than 200mV. 
			// There is a case where setting the Dpot too high causes the DC-DC to drop out.
			int voltageDrop = (adcVoltage - adcVoltageOld);
			if (dpotValue < 253 && voltageDrop >= -200) {
				if (mosfetEnabled < 3) {
					mosfetEnabled++;
				}
				else if (mosfetEnabled >= 3) {
					dpotValue++;
					mosfetEnabled = 0;
				}
			}
			else {
				inCalibration = false;
				calibrationLimit = calibrationCounter-1;
				calibrationCounter = 0;
				calibrated = true;
				PORTD &= ~(1<<LED_ENABLE);
			}
		}
		// Wait until we hit under to 2V
		if (waitForLowVoltage == true && adcVoltage <= 200) {
			waitForLowVoltage = false;
			adcVoltageOld = adcVoltage;
		}
		
		
		// In current section, increase voltage if needed to match current limit
		if (selectedOption == CHANGE_CURRENT) {
			currentLimitCalc = currentLimit / 1.6;
			int currentDifference = currentNumber - currentLimitCalc;
			
			if (currentNumber >= 5 && (dpotValue < 240 || calibrationCounter < calibrationLimit)) {
				// Large differences
				int largeDifference = currentNumber / -10;
				while (currentNumber >= 50 && currentDifference <= largeDifference && (dpotValue < 240 || calibrationCounter < calibrationLimit)) {
					if (calibrated == false) {
						if (dpotValue < 240) { // See which pin is low
							dpotValue++;
						}
					}
					else { // Calibrated
						if (calibrationCounter < calibrationLimit) {
							calibrationCounter++;
						}
					}
					
					update_voltage_dpot_mosfet();
					update_current_limit();
					_delay_ms(10);
					
					voltageNumber = (int) (double) adc_read(7) * (double) 0.00107421875 * (double) 14.7 * (double) 100;
					_delay_ms(1);
					currentNumber = (int) (double) adc_read(6) * (double) 0.00107421875 * (double) 2 * (double) 1000;
					
					currentLimitCalc = currentLimit / 1.6;
					currentDifference = currentNumber - currentLimitCalc;
					display_screen();
				}
				
				// Small differences
				if (currentNumber <= 500 && currentDifference <= -3) {
					if (calibrated == false) {
						if (dpotValue < 240) { // See which pin is low
							dpotValue++;
						}
					}
					else { // Calibrated
						if (calibrationCounter < calibrationLimit) {
							calibrationCounter++;
						}
					}
				}
			}
		}
		
		// Over current limit, lower voltage
		if (currentNumber >= 5) {
			currentLimitCalc = currentLimit / 1.6;
			int currentDifference = currentNumber - currentLimitCalc;
			if (currentDifference >= -2) {
				currentLimited = true;
			}
			
			// Large differences
			int largeDifference = currentNumber / 10;
			while (currentNumber >= 50 && currentDifference >= largeDifference && (dpotValue > 50 || calibrationCounter > 0)) {	
				if (calibrated == false) {
					if (dpotValue > 50) { // See which pin is low
						dpotValue--;
					}
				}
				else { // Calibrated
					if (calibrationCounter > 0) {
						calibrationCounter--;
					}
				}
				
				update_voltage_dpot_mosfet();
				update_current_limit();
				_delay_ms(10);
				
				voltageNumber = (int) (double) adc_read(7) * (double) 0.00107421875 * (double) 14.7 * (double) 100;
				_delay_ms(1);
				currentNumber = (int) (double) adc_read(6) * (double) 0.00107421875 * (double) 2 * (double) 1000;
				
				currentLimitCalc = currentLimit / 1.6;
				currentDifference = currentNumber - currentLimitCalc;
				display_screen();
			}
			
			// Small differences
			if (currentDifference >= -1) {
				if (calibrated == false) {
					if (dpotValue > 50) { // See which pin is low
						dpotValue--;
					}
				}
				else { // Calibrated
					if (calibrationCounter > 0) {
						calibrationCounter--;
					}
				}
			}
			
			// Current no longer limited?
			if (currentDifference < -2) {
				currentLimited = false;
			}
		}
		else {
			currentLimited = false;
		}
		
		
		display_screen();
		_delay_ms(5);
	}
}

EMPTY_INTERRUPT(ADC_vect);

ISR(PCINT2_vect) {
	_delay_ms(1);
	
	if (PIND & (1<<SW_ENABLE)) { // If high
		inCalibration = true;
		waitForLowVoltage = true;
		calibrated = false;
		dpotValue = 50; // Starting value
		dpotValueOld = 0;
		mosfetEnabled = 0;
	}
	
	while (PIND & (1<<SW_ENABLE)); // Wait until low
	_delay_ms(5);
}


// Timer1 compare
ISR(TIMER1_COMPA_vect) {
	TCCR1B = 0;
	encoderTimeout = true;
}

// Pin interrupt on rotary encoder pins
ISR(PCINT0_vect) {
	_delay_ms(1);
	
	if (PINB & (1<<SW_SELECT)) { // Switch select
		if (selectedOption == CHANGE_VOLTAGE) {
			selectedOption = CHANGE_CURRENT;
		}
		else {
			selectedOption = CHANGE_VOLTAGE;
		}
		
		while (PINB & (1<<SW_SELECT)); // Wait until low
		_delay_ms(5);
	}
	else {
		// Waiting for either pin to go low
		if (encoderWaitforlow == true && !((PINB & (1<<PB1)) && (PINB & (1<<PB0)))) {
			
			// Setting Voltage and make sure not in current limit if wanting to increase voltage
			if (selectedOption == CHANGE_VOLTAGE) {
				if (calibrated == false) {
					if ((PINB & (1<<PB0)) && dpotValue > 50) { // See which pin is low
						dpotValue--;
					}
					else if ((PINB & (1<<PB1)) && dpotValue < 250 && currentLimited == false) {
						dpotValue++;
					}
				}
				else { // Calibrated
					if ((PINB & (1<<PB0)) && calibrationCounter > 0) { // See which pin is low
						calibrationCounter--;
					}
					else if ((PINB & (1<<PB1)) && calibrationCounter < calibrationLimit && currentLimited == false) {
						calibrationCounter++;
					}
				}
			}
			else { // Setting Current
				if ((PINB & (1<<PB0)) && currentLimit > 10) { // See which pin is low
					currentLimit -= 5;
				}
				else if ((PINB & (1<<PB1)) && currentLimit < 4000) {
					currentLimit += 5;
				}
				settingCurrent = 1;
			}
			
			// Wait for 3ms
			OCR1A = 562;
			TCNT1 = 0;
			TCCR1B |= (1<<CS11) | (1<<CS10);
			encoderWaitforlow = false;
			encoderTimeout = false;
		}
		else if (encoderWaitforlow == false && encoderTimeout == true) { // After the 3ms
			if (!((PINB & (1<<PB1)) && (PINB & (1<<PB0)))) { // If either pin is still low, wait another 500uS
				OCR1A = 93;
				TCNT1 = 0;
				TCCR1B |= (1<<CS11) | (1<<CS10);
			}
			else {
				encoderWaitforlow = true;
			}
		}
	}
}
